@@ -28,6 +28,7 @@
 #include <stdlib.h>
 #include <string.h>
 #include <errno.h>
+#include <inttypes.h>
 
 #include <glib.h>
 
@@ -60,50 +61,51 @@ static void isi_clear_topics(struct ofono_cbs *cbs,
 	CALLBACK_WITH_FAILURE(cb, data);
 }
 
-static void routing_ntf_cb(GIsiClient *client, const void *restrict data,
-				size_t len, uint16_t object, void *opaque)
+static void routing_ntf_cb(GIsiClient *client,
+				const void *restrict data, size_t len,
+				uint16_t object, void *opaque)
 {
 	const unsigned char *msg = data;
 	struct ofono_cbs *cbs = opaque;
 
-	DBG("");
-
 	if (!msg || len < 3 || msg[0] != SMS_GSM_CB_ROUTING_NTF)
 		return;
 
-	ofono_cbs_notify(cbs, msg+3, len-3);
+	ofono_cbs_notify(cbs, msg+5, len-5);
 }
 
-static bool routing_resp_cb(GIsiClient *client, const void *restrict data,
-				size_t len, uint16_t object, void *opaque)
+static gboolean routing_resp_cb(GIsiClient *client,
+				const void *restrict data, size_t len,
+				uint16_t object, void *opaque)
 {
 	const unsigned char *msg = data;
 	struct ofono_cbs *cbs = opaque;
-	const char *debug = NULL;
 
 	if (!msg) {
 		DBG("ISI client error: %d", g_isi_client_error(client));
-		return true;
+		return TRUE;
 	}
 
 	if (len < 3 || msg[0] != SMS_GSM_CB_ROUTING_RESP)
-		return false;
+		return FALSE;
 
 	if (msg[1] != SMS_OK) {
-		DBG("Request failed: 0x%02X (%s).\n\n  Unable to bootstrap CBS"
-			" routing.\n  It appears some other component is"
-			" already\n  registered as the CBS routing endpoint.\n "
-			" As a consequence, receiving CBSs is NOT going"
-			" to work.\n\n", msg[1], sms_isi_cause_name(msg[1]));
-		return true;
+		if (msg[1] == SMS_ERR_PP_RESERVED)
+			DBG("Request failed: 0x%02"PRIx8" (%s).\n\n  "
+				"Unable to bootstrap CBS routing.\n  "
+				"It appears some other component is "
+				"already\n  registered as the CBS "
+				"routing endpoint.\n  As a consequence, "
+				"receiving CBSs is NOT going to work.\n\n",
+				msg[1], sms_isi_cause_name(msg[1]));
+		return TRUE;
 	}
 
-	debug = getenv("OFONO_ISI_DEBUG");
-	if (debug && (strcmp(debug, "all") == 0 || strcmp(debug, "cbs") == 0))
-		g_isi_client_set_debug(client, sms_debug, NULL);
+	g_isi_subscribe(client, SMS_GSM_CB_ROUTING_NTF, routing_ntf_cb,
+			cbs);
 
 	ofono_cbs_register(cbs);
-	return true;
+	return TRUE;
 }
 
 static int isi_cbs_probe(struct ofono_cbs *cbs, unsigned int vendor,
@@ -111,6 +113,7 @@ static int isi_cbs_probe(struct ofono_cbs *cbs, unsigned int vendor,
 {
 	GIsiModem *idx = user;
 	struct cbs_data *cd = g_try_new0(struct cbs_data, 1);
+	const char *debug = NULL;
 
 	unsigned char msg[] = {
 		SMS_GSM_CB_ROUTING_REQ,
@@ -134,12 +137,14 @@ static int isi_cbs_probe(struct ofono_cbs *cbs, unsigned int vendor,
 
 	ofono_cbs_set_data(cbs, cd);
 
+	debug = getenv("OFONO_ISI_DEBUG");
+	if (debug && (strcmp(debug, "all") == 0 || strcmp(debug, "cbs") == 0))
+		g_isi_client_set_debug(cd->client, sms_debug, NULL);
+
 	if (!g_isi_request_make(cd->client, msg, sizeof(msg), CBS_TIMEOUT,
 				routing_resp_cb, cbs))
 		DBG("Failed to set CBS routing.");
 
-	g_isi_subscribe(cd->client, SMS_GSM_CB_ROUTING_NTF, routing_ntf_cb, cbs);
-
 	return 0;
 }
 
@@ -147,10 +152,31 @@ static void isi_cbs_remove(struct ofono_cbs *cbs)
 {
 	struct cbs_data *data = ofono_cbs_get_data(cbs);
 
-	if (data) {
+	uint8_t msg[] = {
+		SMS_GSM_CB_ROUTING_REQ,
+		SMS_ROUTING_RELEASE,
+		SMS_GSM_ROUTING_MODE_ALL,
+		SMS_CB_NOT_ALLOWED_IDS_LIST,
+		0x00,  /* Subject count */
+		0x00,  /* Language count */
+		0x00,  /* CB range */
+		0x00,  /* Subject list MSBS */
+		0x00,  /* Subject list LSBS */
+		0x00   /* Languages */
+	};
+
+	if (!data)
+		return;
+
+	if (data->client) {
+		/* Send a promiscuous routing release, so as not to
+		 * hog resources unnecessarily after being removed */
+		g_isi_request_make(data->client, msg, sizeof(msg),
+					CBS_TIMEOUT, NULL, NULL);
 		g_isi_client_destroy(data->client);
-		g_free(data);
 	}
+
+	g_free(data);
 }
 
 static struct ofono_cbs_driver driver = {
