@@ -30,7 +30,6 @@ void __ofono_exit();
 int __ofono_manager_init();
 void __ofono_manager_cleanup();
 
-const char **__ofono_modem_get_list();
 void __ofono_modem_shutdown();
 
 #include <ofono/log.h>
@@ -90,9 +89,14 @@ void __ofono_plugin_cleanup(void);
 
 #include <ofono/modem.h>
 
+typedef void (*ofono_modem_foreach_func)(struct ofono_modem *modem, void *data);
+void __ofono_modem_foreach(ofono_modem_foreach_func cb, void *userdata);
+
 unsigned int __ofono_modem_callid_next(struct ofono_modem *modem);
 void __ofono_modem_callid_hold(struct ofono_modem *modem, int id);
 void __ofono_modem_callid_release(struct ofono_modem *modem, int id);
+void __ofono_modem_append_properties(struct ofono_modem *modem,
+						DBusMessageIter *dict);
 
 struct ofono_atom;
 
@@ -163,9 +167,18 @@ gboolean __ofono_modem_remove_atom_watch(struct ofono_modem *modem,
 void __ofono_atom_free(struct ofono_atom *atom);
 
 #include <ofono/call-barring.h>
+
+gboolean __ofono_call_barring_is_busy(struct ofono_call_barring *cb);
+
 #include <ofono/call-forwarding.h>
+
+gboolean __ofono_call_forwarding_is_busy(struct ofono_call_forwarding *cf);
+
 #include <ofono/call-meter.h>
 #include <ofono/call-settings.h>
+
+gboolean __ofono_call_settings_is_busy(struct ofono_call_settings *cs);
+
 #include <ofono/cbs.h>
 #include <ofono/devinfo.h>
 #include <ofono/phonebook.h>
@@ -205,10 +218,10 @@ enum ofono_sms_submit_flag {
 
 typedef void (*ofono_sms_txq_submit_cb_t)(gboolean ok, void *data);
 
-unsigned int __ofono_sms_txq_submit(struct ofono_sms *sms, GSList *list,
-					unsigned int flags,
-					ofono_sms_txq_submit_cb_t cb,
-					void *data, ofono_destroy_func destroy);
+int __ofono_sms_txq_submit(struct ofono_sms *sms, GSList *list,
+				unsigned int flags, struct ofono_uuid *uuid,
+				ofono_sms_txq_submit_cb_t cb,
+				void *data, ofono_destroy_func destroy);
 
 #include <ofono/sim.h>
 #include <ofono/stk.h>
@@ -245,6 +258,9 @@ typedef gboolean (*ofono_ussd_passwd_cb_t)(const char *sc,
 					const char *old, const char *new,
 					DBusMessage *msg, void *data);
 
+typedef void (*ofono_ussd_request_cb_t)(int error, int dcs,
+					const unsigned char *pdu, int len, void *data);
+
 gboolean __ofono_ussd_ssc_register(struct ofono_ussd *ussd, const char *sc,
 					ofono_ussd_ssc_cb_t cb, void *data,
 					ofono_destroy_func destroy);
@@ -254,6 +270,12 @@ gboolean __ofono_ussd_passwd_register(struct ofono_ussd *ussd, const char *sc,
 					ofono_ussd_passwd_cb_t cb, void *data,
 					ofono_destroy_func destroy);
 void __ofono_ussd_passwd_unregister(struct ofono_ussd *ussd, const char *sc);
+gboolean __ofono_ussd_is_busy(struct ofono_ussd *ussd);
+
+int __ofono_ussd_initiate(struct ofono_ussd *ussd, int dcs,
+			const unsigned char *pdu, int len,
+			ofono_ussd_request_cb_t cb, void *user_data);
+void __ofono_ussd_initiate_cancel(struct ofono_ussd *ussd);
 
 #include <ofono/netreg.h>
 
@@ -283,17 +305,20 @@ void __ofono_history_call_missed(struct ofono_modem *modem,
 				const struct ofono_call *call, time_t when);
 
 void __ofono_history_sms_received(struct ofono_modem *modem,
-					unsigned int msg_id, const char *from,
+					const struct ofono_uuid *uuid,
+					const char *from,
 					const struct tm *remote,
 					const struct tm *local,
 					const char *text);
 
 void __ofono_history_sms_send_pending(struct ofono_modem *modem,
-					unsigned int msg_id, const char *to,
+					const struct ofono_uuid *uuid,
+					const char *to,
 					time_t when, const char *text);
 
 void __ofono_history_sms_send_status(struct ofono_modem *modem,
-					unsigned int msg_id, time_t when,
+					const struct ofono_uuid *uuid,
+					time_t when,
 					enum ofono_history_sms_status status);
 
 #include <ofono/message-waiting.h>
